import traceback

import ipywidgets.widgets as widgets
from ipywidgets import Layout


class IndependentParameterWidget(widgets.VBox):
    """Widget for displaying and controlling independent parameters."""

    def __init__(self, paramobj, escape_obj, checkbox_property="fixed", **kwargs):
        """Initialize parameter widget.

        Args:
            paramobj: Parameter object to control
            escape_obj: Parent escape object
            checkbox_property: Property to toggle ('fixed' or 'trained')
            **kwargs: Additional widget arguments

        Raises:
            ValueError: If checkbox_property is invalid
        """
        widgets.HBox.__init__(self, **kwargs)
        if checkbox_property not in ["fixed", "trained"]:
            raise ValueError("Checkbox property must be either 'fixed' or 'trained'")

        self.paramobj = paramobj
        self.escape_obj = escape_obj
        self.checkbox_property = checkbox_property
        self.children = self.init_children()

    def init_children(self):
        paramobj = self.paramobj
        pmin, pmax = paramobj.get_unscaled_limits()
        pval = paramobj.get_unscaled_value()
        step = (pmax - pmin) / 100.0
        pval = paramobj.get_unscaled_value()
        self._disabled = False
        description = ""
        un = paramobj.units
        units = f"[{un}]" if un else ""
        if paramobj.scale != 1.0:
            description = "{0} x{1:.4g}{2}:".format(
                paramobj.name, paramobj.scale, units
            )
        else:
            description = "{0}{1}:".format(paramobj.name, units)
        self.description = description
        try:
            slider = widgets.FloatSlider(
                value=pval,
                min=pmin,
                max=pmax,
                step=step,
                readout=True,
                readout_format=".4g",
                continuous_update=False,
                description="",
                disabled=False,
                layout=Layout(width="90%", height="20px"),
            )
            checkbox_property = self.checkbox_property
            if checkbox_property == "fixed":
                chkbox = widgets.Checkbox(
                    value=not paramobj.fixed,
                    description=description,
                    description_tooltip="Fit/Fixed",
                    indent=False,
                    layout=Layout(width="90%", height="15px"),
                )
            elif checkbox_property == "trained":
                chkbox = widgets.Checkbox(
                    value=paramobj.trained,
                    description=description,
                    description_tooltip="Trained/not Trained",
                    indent=False,
                    layout=Layout(width="90%", height="15px"),
                )
            slider.observe(self.update_value, names="value")
            chkbox.observe(self.update_property, names="value")
            self.slider = slider
            self.chkbox = chkbox
            return [
                chkbox,
                slider,
            ]
        except:
            traceback.print_exc()
            return []

    def update(self, *args):
        self.update_value(*args)

    def update_value(self, *args):
        """Update parameter value from slider.

        Maintains parameter feasibility by reverting invalid values.
        """
        try:
            old_value = self.paramobj.value
            value = self.slider.get_interact_value()
            if value == old_value:
                return

            self.paramobj.value = value
            if not self.escape_obj.is_feasible:
                self.paramobj.value = old_value
                self.slider.value = old_value
        except Exception as e:
            print(f"Error updating parameter value: {e}")

    def pull_value(self):
        parobj = self.paramobj
        v = parobj.get_unscaled_value()
        self.slider.value = v
        checkbox_property = self.checkbox_property
        chkbox = self.chkbox
        if checkbox_property == "fixed":
            chkbox.value = not parobj.fixed
        elif checkbox_property == "trained":
            chkbox.value = parobj.trained

    def update_property(self, *args):
        value = self.chkbox.get_interact_value()
        checkbox_property = self.checkbox_property
        if checkbox_property == "fixed":
            self.paramobj.fixed = not value
        elif checkbox_property == "trained":
            self.paramobj.trained = value

    @property
    def disabled(self):
        return self._disabled

    @disabled.setter
    def disabled(self, val):
        self._disabled = val
        for c in self.children:
            c.disabled = val

    def _repr_mimebundle_(self, **kwargs):
        self.update()
        return super()._repr_mimebundle_(**kwargs)


class DependentParameterWidget(widgets.VBox):

    label_tpl = "Result: {0:.4g}"

    def __init__(self, paramobj, checkbox_property="fixed", **kwargs):
        widgets.VBox.__init__(self, _dom_classes=["widget-interact"])
        self.paramobj = paramobj
        self.ppn = ParametersPanel(paramobj, checkbox_property=checkbox_property)
        self.lbl = widgets.Label("")
        for par in self.ppn.children:
            par.slider.observe(self.update, names="value")
        self.children = [self.ppn, self.lbl]

    # Callback function
    def update(self, *args):
        self.lbl.value = self.label_tpl.format(self.paramobj.value)


class ParametersPanel(widgets.VBox):

    def __init__(self, escape_obj, checkbox_property="fixed", **kwargs):
        widgets.VBox.__init__(self, _dom_classes=["widget-interact"], **kwargs)
        if not isinstance(escape_obj, (list, tuple)):
            escape_obj = [escape_obj]
        params_ids = set()
        children = []
        for escobj in escape_obj:
            params = escobj.parameters
            children += [
                IndependentParameterWidget(
                    p,
                    escobj,
                    checkbox_property=checkbox_property,
                    layout=Layout(
                        max_height="50px",
                        min_height="50px",
                        overflow_x="hidden",
                        overflow_y="hidden",
                    ),
                )
                for p in params
                if p.id not in params_ids
            ]
            params_ids.update([p.id for p in params])

        self.children = children
        self._disabled = False

    @property
    def disabled(self):
        return self._disabled

    @disabled.setter
    def disabled(self, val):
        self._disabled = val
        for c in self.children:
            c.disabled = val
